/*=============================================================================
gdmalloc - Memory leak detection helper utility
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/

/*-----------------------------------------------------------------------------
GDmalloc is intended to be as free as possible. You do not need to do anything
else than passing -lgdmalloc -rdynamic (and probably -g ;-) to gcc for your
executable. No CFLAGS, no define, no include, no alternate names. Nothing.
Even more: you can set the environment variable LD_PRELOAD to ./libgdmalloc.so
to memory-debug about everything, even binary-only software (which does NOT
call internal, statically-linekd malloc, calloc, realloc and free, if
applicable.)

And thus, GDmalloc only compiles with GCC.

Options are set via the environment variable GDMALLOC. Current flags:
  C     do not use ANSI color
-----------------------------------------------------------------------------*/
#define _GNU_SOURCE 1
#include <sys/types.h>
#include <dlfcn.h>
#include <malloc.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
//define SBT_LEVEL   /* defined in stack_backtrace() */
#define SBT_SIZE 128 /* maximum buffer size for stack backtrace dumps */
#define _weak_alias(name, aname) \
  extern __typeof(name) aname __attribute__((weak, alias(#name)));
#define _strong_alias(name, aname) \
  extern __typeof(name) aname __attribute__((alias(#name)));

typedef unsigned char byte;
typedef unsigned long Ulong;

// Function prototypes
extern void *__libc_malloc(size_t);
extern void *__libc_calloc(size_t, size_t);
extern void *__libc_realloc(void *, size_t);
extern void __libc_free(void *);

extern void _GDM_init(void *) __attribute__((constructor));
extern void _GDM_deinit(void) __attribute__((destructor));
extern void *_GDM_malloc_hook(size_t, const void *);
extern void *_GDM_realloc_hook(void *, size_t, const void *);
extern void _GDM_free_hook(void *, const void *);

static void dump_unfreed(void);
static char *escape(const char *, char *, size_t);
static int flag(char);
static void hooks_deploy(void);
static void hooks_restore(void);
static int print(char, char *, ...);
static const char *rsymbol(const void *);
static char *stack_backtrace(char *, size_t);
static int value_cmp(void *, void *);

// Variables
static void *(*orig_malloc_hook)(size_t, const void *) = NULL;
static void *(*orig_realloc_hook)(void *, size_t, const void *) = NULL;
static void (*orig_free_hook)(void *, const void *) = NULL;
static pthread_mutex_t ext_lock = PTHREAD_MUTEX_INITIALIZER;
static struct HXbtree *AO = NULL;
static char *flags = NULL;
static int _GDM_inited = 0;

//-----------------------------------------------------------------------------
// Internal functions
void _GDM_init(void *caller) {
  if(_GDM_inited) {
    print(0, "o_O inter-thread access to _GDM_init, not good!\n");
    return;
  }

  pthread_mutex_init(&ext_lock, NULL);
  pthread_mutex_lock(&ext_lock);
  if(_GDM_inited++) { return; }

  flags = getenv("GDMALLOC");
  AO = HXbtree_init(HXBT_CPTR | HXBT_HASH | HXBT_LKEY | HXBT_PCMP, value_cmp);
  print(0, "GDmalloc hooks initialized\n");

  orig_malloc_hook = __malloc_hook;
  orig_realloc_hook = __realloc_hook;
  orig_free_hook = __free_hook;
  pthread_mutex_unlock(&ext_lock);
  hooks_deploy();
  return;
}

void _GDM_deinit(void) {
  hooks_restore();
  print(0, "GDmalloc hooks deinitialized\n");
  if(AO->itemcount > 0) { dump_unfreed(); }
  HXbtree_free(AO);
  pthread_mutex_destroy(&ext_lock);
  return;
}

//-----------------------------------------------------------------------------
// Re-wrappers
void *malloc(size_t s) { return __libc_malloc(s); }
void *calloc(size_t n, size_t s) { return __libc_calloc(n, s); }
void *realloc(void *ptr, size_t s) { return __libc_realloc(ptr, s); }
void free(void *ptr) { __libc_free(ptr); return; }

//-----------------------------------------------------------------------------
// Libc hooks
void *_GDM_malloc_hook(size_t s, const void *caller) {
  void *rv;

  pthread_mutex_lock(&ext_lock);
  hooks_restore();

  if((rv = __libc_malloc(s)) != NULL) {
    char buf[SBT_SIZE];
    HXbtree_add(AO, rv, stack_backtrace(buf, SBT_SIZE));
  }
  print('v', "[objs=%5ld] 0x%08lX = malloc(s=0x%08lX)\n",
   AO->itemcount, (Ulong)rv, (Ulong)s);

  hooks_deploy();
  pthread_mutex_unlock(&ext_lock);
  return rv;
}

void *_GDM_realloc_hook(void *ptr, size_t s, const void *caller) {
  void *rv;

  pthread_mutex_lock(&ext_lock);
  hooks_restore();

  rv = __libc_realloc(ptr, s);
  if(ptr != rv) {
    if(ptr != NULL) { HXbtree_del(AO, ptr); }
    if(rv != NULL) {
      char buf[SBT_SIZE];
      HXbtree_add(AO, rv, stack_backtrace(buf, SBT_SIZE));
    }
  }

  print('v', "[objs=%5ld] 0x%08lX = realloc(ptr=0x%08lX, s=0x%08lX)\n",
   AO->itemcount, (Ulong)rv, (Ulong)ptr, (Ulong)s);

  hooks_deploy();
  pthread_mutex_unlock(&ext_lock);
  return rv;
}

void _GDM_free_hook(void *ptr, const void *caller) {
  pthread_mutex_lock(&ext_lock);
  hooks_restore();
  __libc_free(ptr);

  HXbtree_del(AO, ptr);
  print('v', "[objs=%5ld] free(ptr=0x%08lX)\n",
   AO->itemcount, (Ulong)ptr);

  hooks_deploy();
  pthread_mutex_unlock(&ext_lock);
  return;
}

//-----------------------------------------------------------------------------
// Internal functions
static void dump_unfreed(void) {
  struct HXbtree_trav1 *travp = NULL;
  while((travp = HXbtree_trav1(AO, travp)) != NULL) {
    print(0, "Unfreed object 0x%08lX (BT: %s)\n",
     (Ulong)travp->node->key, travp->node->data);
  }
  return;
}

static char *escape(const char *src, char *buf, size_t len) {
  char ibuf[512];

  while(*src != '\0') {
    if(*src < ' ' || *src >= 0x7F) {
      snprintf(ibuf, 512, "\\x%02X", *src);
      HX_strlcat(buf, ibuf, len);
      ++src;
      continue;
    }

    snprintf(ibuf, 512, "%c", *src);
    HX_strlcat(buf, ibuf, len);
    ++src;
  }

  return buf;
}

static int flag(char f) {
  if(f == 0 || (flags != NULL && strchr(flags, f) != NULL)) { return 1; }
  return 0;
}

static void hooks_deploy(void) {
  __malloc_hook = _GDM_malloc_hook;
  __realloc_hook = _GDM_realloc_hook;
  __free_hook = _GDM_free_hook;
  return;
}

static void hooks_restore(void) {
  __malloc_hook = orig_malloc_hook;
  __realloc_hook = orig_realloc_hook;
  __free_hook = orig_free_hook;
  return;
}

static int print(char f, char *fmt, ...) {
  int rv = 0;
  va_list argp;
  va_start(argp, fmt);
  if(flag(f)) {
    if(!flag('C')) { fprintf(stderr, "\e[35mGDM\e[0;1m]\e[0m "); }
    else { fprintf(stderr, /* [ */ "GDM] "); }
    rv = vfprintf(stderr, fmt, argp);
  }
  va_end(argp);
  return rv;
}

static const char *rsymbol(const void *addr) {
  Dl_info q;
  if(!dladdr(addr, &q) || q.dli_sname == NULL) { return NULL; }
  return q.dli_sname;
}

static char *stack_backtrace(char *buf, size_t len) {
#define SBT_LEVEL 5 /* maximum number of stack frames to dump */
  struct {
    const void *addr;
    const char *sym;
  } sym[SBT_LEVEL] = {
    {__builtin_return_address(3)},
    {__builtin_return_address(4)},
    {__builtin_return_address(5)},
    {__builtin_return_address(6)},
    {__builtin_return_address(7)},
  };

  int i;
  char ibuf[512];
  *buf = '\0';

  for(i = 0; i < SBT_LEVEL; ++i) {
    sym[i].sym = rsymbol(sym[i].addr);
    if(sym[i].sym == NULL) {
      snprintf(ibuf, 512, "0x%08lX", (unsigned long)sym[i].addr);
    } else {
      snprintf(ibuf, 512, "%s", sym[i].sym);
    }

    HX_strlcat(buf, ibuf, len);

    if(sym[i].sym != NULL && strcmp(sym[i].sym, "main") == 0) { break; }
    if(i != SBT_LEVEL - 1 || strlen(buf) + 5 > len) {
      HX_strlcat(buf, " > ", len);
    }
  }

  return buf;
#undef SBT_LEVEL
}

static int value_cmp(void *pa, void *pb) {
  return (pa > pb) - (pa < pb);
}

//-----------------------------------------------------------------------------
// Redirect any malloc() call to us, then __libc_malloc()
_strong_alias(malloc,  _GDM_malloc);
_strong_alias(calloc,  _GDM_calloc);
_strong_alias(realloc, _GDM_realloc);
_strong_alias(free,    _GDM_free);

// External-namespace clean symbols, for debugging purposes
_strong_alias(dump_unfreed,    _GDM_dump_unfreed);
_strong_alias(escape,          _GDM_escape);
_strong_alias(flag,            _GDM_flag);
_strong_alias(hooks_deploy,    _GDM_hooks_deploy);
_strong_alias(hooks_restore,   _GDM_hooks_restore);
_strong_alias(print,           _GDM_print);
_strong_alias(rsymbol,         _GDM_rsymbol);
_strong_alias(stack_backtrace, _GDM_stack_backtrace);
_strong_alias(value_cmp,       _GDM_value_cmp);

//==[ End of file ]============================================================
