/*=============================================================================
daxtraq - User Management UI
ptui.c - PT Console UI (plaintext)
  Copyright (C) Jan Engelhardt <jengelh [at] linux01 gwdg de>, 1999 - 2003
  -- License restrictions apply (Dual FAL/GPL2).
  -- For details see doc/FAL.txt and doc/GPL2.txt.
=============================================================================*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libHX.h>
#include <accdb.h>
#include "dxt/global.h"
#include "dxt/mm.h"

static void apply_ds(void);
static void cleanup_storage_addlist(void);
static void cleanup_storage_dellist(void);
static void free_addlist_entry(struct ds_entry *);
static inline void handle_user(struct accdb_user *, size_t *);
static int get_dgrp(struct accdb_group *);
static int read_input(void);

static struct HXbtree *storage_addlist = NULL, *storage_dellist = NULL;
static struct accdb_module *mm_output = NULL;

//-----------------------------------------------------------------------------
void Mainmenu_TX(void) {
  mm_register_ds();

  mm_output = accdb_load("shadow");
  mm_output->open(&mm_output->state);

  printf("\n");
  apply_ds();

  mm_unregister_ds();
  mm_output->close(mm_output->state);
  accdb_unload(mm_output);
  return;
}

static void apply_ds(void) {
  struct HXbtree *lnlist;
  struct accdb_group dgroup;

  cleanup_storage_addlist();
  cleanup_storage_dellist();

  if(!get_dgrp(&dgroup)) { return; }
  if(!read_input()) { return; }

  { /* And this is now the comparison algorithm, (half of) the core. While
    traversing the users, we mark used login names by putting them in an array
    (lnlist). Login names of users which are about to be deleted are also
    pushed there, so they do not get reassigned right now, but at the next
    run-through. */
    size_t users_grp = 0, users_keep = 0;
    struct accdb_user passwd;

    lnlist = HXbtree_init(HXBT_CPTR | HXBT_PCMP, strcmp);
    storage_dellist = HXbtree_init(HXBT_PCMP, strcmp);

    while(mm_output->usertrav(mm_output->state, &passwd)) {
      HXbtree_add(lnlist, passwd.lname);

      if((dgroup.gid == -1 || dgroup.gid == passwd.gid) &&
       (dgroup.gname == NULL || passwd.lname == NULL ||
       strcmp(dgroup.gname, passwd.lname) == 0)) {
        // Users we look for (only specified group)
        ++users_grp;
        handle_user(&passwd, &users_keep);
      }
    }

    printf("%u group member%s found\n", users_grp, users_grp != 1 ? "s" : "");
    printf("\t %u to keep\n", users_keep);
    printf("\t %u to delete\n", storage_dellist->itemcount);
    printf("%u to add\n", storage_addlist->itemcount);
  }

  { // Here comes the fixup loop (other part of the core).
    struct HXbtree_trav1 *travp = NULL;

    while((travp = HXbtree_trav1(storage_addlist, travp)) != NULL) {
      struct ds_entry *entry = travp->node->data;
      char buf[16];
      int i = 0;
      strncpy(buf, entry->lname, 16);

      /* If the fixup'ed login name with the unique number already exists,
      bump the unique number. */
      while(HXbtree_find(lnlist, buf) != NULL) {
        fixup_login_name(entry->lname, ++i, buf, 16);
      }

      HX_strclone(&entry->lname, buf);
      HXbtree_add(lnlist, buf);
    }
  }

  HXbtree_free(lnlist);

  // Adding the users to the system database
  if(storage_addlist->itemcount > 0) {
    struct HXbtree_trav1 *travp = NULL;

    while((travp = HXbtree_trav1(storage_addlist, travp)) != NULL) {
      struct ds_entry *in = travp->node->data;
      struct accdb_user out;

      memset(&out, '\0', sizeof(struct accdb_user));
      out.lname = in->lname;
      out.uid = -1;
      out.gid = dgroup.gid;
      out.group = dgroup.gname;
      out.gecos = travp->node->key;
      out.home = NULL;
      out.shell = NULL;
    }
  }

  // Deleting the old users
  if(storage_dellist->itemcount > 0) {
    struct HXbtree_trav1 *travp = NULL;

    while((travp = HXbtree_trav1(storage_dellist, travp)) != NULL) {
      printf("delete %s\n", (char *)travp->node->key);
    }
  }

  cleanup_storage_addlist();
  cleanup_storage_dellist();
  return;
}

static void cleanup_storage_addlist(void) {
  if(storage_addlist != NULL) {
    struct HXbtree_trav1 *travp = NULL;
    while((travp = HXbtree_trav1(storage_addlist, travp)) != NULL) {
      free_addlist_entry(travp->node->data);
    }
    HXbtree_free(storage_addlist);
    storage_addlist = NULL;
  }
  return;
}

static void cleanup_storage_dellist(void) {
  if(storage_dellist != NULL) {
    HXbtree_free(storage_dellist);
    storage_dellist = NULL;
  }
  return;
}

static void free_addlist_entry(struct ds_entry *e) {
  if(e == NULL) { return; }
  free(e->vname);
  if(e->nname != NULL) { free(e->nname); }
  if(e->xuid_in != NULL) { free(e->xuid_in); }
  if(e->sgroup != NULL) { free(e->sgroup); }
  free(e->lname);
  if(e->xuid != NULL) { free(e->xuid); }
  free(e->rname);
  return;
}

static int get_dgrp(struct accdb_group *grp) {
  /* Here, we request a group to which compare the data source against.
  Usually, a group is always the same for a particular input set otherwise
  strange things may happen. You can either enter the group name or a GID.
  (Once there is a non-digit in your input, it is treated as a group name.) */
  char buf[MAXLNLEN];

  printf("Destianation group> ");
  fgets(buf, MAXLNLEN, stdin);
  HX_chomp(buf);
  if(strlen(buf) < 1) { return 0; }

  if(only_digits(buf)) {
    grp->gid = strtol(buf, NULL, 0);
    grp->gname = NULL;
  } else {
    grp->gid = -1;
    grp->gname = buf;
  }

  if(!mm_output->groupinfo(mm_output->state, grp, grp)) {
    printf("Group \"%s\" not found\n", buf);
    return 0;
  }

  return 1;
}

static inline void handle_user(struct accdb_user *p, size_t *ukeep) {
  char buf[MAXLNLEN], *brkp;

  memset(buf, '\0', 512);
  strncpy(buf, p->gecos, 511);

  if((brkp = strchr(buf, ',')) != NULL && (brkp = strchr(buf, ',')) != NULL) {
    *brkp = '\0';
  }

  if(HXbtree_find(storage_addlist, buf) != NULL) {
    ++*ukeep;
    free_addlist_entry(HXbtree_get(storage_addlist, buf));
    HXbtree_del(storage_addlist, buf);
    // what remains in storage_addlist is the users to add
  } else {
    // not anymore in data source
    HXbtree_add(storage_dellist, p->lname);
  }

  return;
}

static int read_input(void) {
  /* In this part, we stuff all users from the data source into a hash, called
  storage_addlist. Some elements of that hash will further be changed until it
  is ready for running through diff_and_patch(). */
  struct ds_entry *bucket;
  struct ds_module *imtrav = mm_input;
  char buf[MAXLNLEN];

  printf("FIND CHANGES; Input file> ");
  fgets(buf, MAXLNLEN, stdin);
  HX_chomp(buf);

  /* Choose which Data Source module to use. The first one that reports >0
  supports the input data's file format and is therefore used. */
  while(imtrav != NULL) {
    if(imtrav->open(buf, &mm_input->state) > 0) { break; }
    imtrav = imtrav->Next;
  }

  if(imtrav == NULL) { return 0; } // did not find a suitable one

  /* This loop fills the storage_addlist hash by generating a suited
  (pre-stage) login name and converts the bday (.xuid) into Daxtraq's internal
  format.

  Pre-stage login names: these login names may already be used by the system
  and must first be fixed up (i.e. by adding a unique postcount number), so
  that there is no double login name in the user database. */

  storage_addlist = HXbtree_init(HXBT_HASH);

  while(1) {
    bucket = malloc(sizeof(struct ds_entry));
    if(imtrav->read(imtrav->state, bucket) <= 0) {
      free(bucket);
      break;
    }

    bucket->lname = make_login_name(bucket->nname, bucket->vname);
    bucket->xuid = date_scramble(bucket->xuid_in);

    if(bucket->nname == NULL) {
      snprintf(buf, MAXLNLEN, "%s", bucket->vname);
      bucket->rname = HX_strdup(buf);
      snprintf(buf, MAXLNLEN, "%s,%s", bucket->vname, bucket->xuid);
    } else {
      snprintf(buf, MAXLNLEN, "%s %s", bucket->vname, bucket->nname);
      bucket->rname = HX_strdup(buf);
      snprintf(buf, MAXLNLEN, "%s %s,%s", bucket->vname, bucket->nname,
       bucket->xuid);
    }

    HXbtree_add(storage_addlist, buf, bucket);
  }

  imtrav->close(imtrav->state);
  return 1;
}

//==[ End of file ]============================================================
